-- Emilio Francesquini <e.francesquini@ufabc.edu.br>
-- CC-BY-SA
-- 09/2019

module Main where

import Prelude hiding (getLine, putStrLn)
import Data.Bits
import Control.Monad.State


-- Exercício 1
--
-- Considere um tipo descrevendo Pokémons que só podem atacar ou
-- defender, o ataque/defesa pode ser descrito por diversos tipos:
-- numérico descrevendo a força, string descrevendo o efeito, tuplas
-- descrevendo ambos, etc...
--
-- Escreva a instância de Functor para esse tipo.

data Pokemon a = ATK a | DEF a | AD a deriving (Show, Eq)

instance Functor Pokemon where
  fmap = undefined


toAD :: Pokemon a -> Pokemon a
toAD (ATK x) = AD x
toAD (DEF x) = AD x
toAD p = p

instance Applicative Pokemon where
  pure = AD
  AD f <*> p = fmap f p
  ATK f <*> ATK y = ATK (f y)
  ATK f <*> p  = toAD (fmap f p)
  DEF f <*> DEF y = DEF (f y)
  DEF f <*> p  = toAD (fmap f p)

-- Exercício 2
--
-- Escreva a instância de Monads para o tipo Pokémon.

instance Monad Pokemon where
  return = undefined
  (>>=) = undefined


-- Exercício 3
--
-- Escreva a função getLine2, baseando-se na função getLine
-- abaixo. Para isto, reescreva o código contido no else como
-- Applicative.

getLine :: IO String
getLine = do
  x <- getChar
  if x == '\n' then
    return []
  else do
    xs <- getLine
    return (x:xs)

getLine2 :: IO String
getLine2 = undefined


-- Exercício 4
--
-- Reescreva a função putStrLn, como putStrLn2, usando Applicative.

putStrLn :: String -> IO ()
putStrLn xs = do putStr xs
                 putChar '\n'

putStrLn2 :: String -> IO ()
putStrLn2 = undefined


-- Exercício 5
--
-- Abaixo apresentamos um exemplo de uso do State Monad para geração
-- de sequencias de números aleatórios. Baseando-se no exemplo de uso
-- abaixo, implemente as funções monadicNextPrime, monadicNextPrime,
-- monadicNextNPrimes que geram uma sequência crescente de números
-- primos.

-- Este tipo guarda o estado do gerador de números aleatórios.
type RNG = (Int, Int)


-- Dada uma semente devolve um gerador de números aleatórios
simpleRandomInit :: Int -> RNG
simpleRandomInit seed =
  (z, w)
  where
    n1 = (seed * 48947) `mod` 4294967296;
    z = if n1 /= 0 then n1 else 362436069
    n2 = (seed * 104623) `mod` 4294967296;
    w = if n2 /= 0 then n2 else 521288629


-- Este simples gerador de números (pseudo)aleatórios é baseado
-- naquele proposto por George Marsaglia (MWC - multiply with
-- carry). A implementação abaixo é uma adaptação daquela feita por
-- John D. Cook.
nextInt :: RNG -> (Int, RNG)
nextInt (z, w) =
  (u, (z', w'))
  where
    z' = 36969 * (z .&. 65535) + (z `shiftR` 16)
    w' = 18000 * (w .&. 65535) + (w `shiftR` 16);
    u  = (z `shiftL` 16) + w;


---------------------------
-- Utilização aleatórios --
---------------------------

-- Dado um gerador devolve uma tupla com o próximo número aleatório e
-- o estado atualizado do gerador
pureNextInt :: RNG -> (Int, RNG)
pureNextInt = nextInt

-- Dado um gerador devolve uma tupla com 3 números aleatórios e o
-- estado atualizado do gerador
pureNext3 :: RNG -> ([Int], RNG)
pureNext3 s0 =
  let
    (v0, s1) = pureNextInt s0 -- precisa carregar o estado
    (v1, s2) = pureNextInt s1 -- entre cada uma das chamadas
    (v2, s3) = pureNextInt s2 in
    ([v0, v1, v2], s3)

-- Dado um n, inteiro, e um gerador devolve uma tupla com uma lista
-- contendo n números aleatórios e o estado atualizado do gerador
pureNextN :: Int -> RNG -> ([Int], RNG)
pureNextN 0 s0 = ([], s0)
pureNextN n s0 = (x : xs, sn)
  where
    -- Note que é necessário carregar o estado na mão
    (x, s1) = pureNextInt s0
    (xs, sn) = pureNextN (n - 1) s1


-----------------------------------------
-- Utilização aleatórios - State Monad --
-----------------------------------------

-- Versão monádica do gerador de um número aleatório
monadicNextInt :: State RNG Int
monadicNextInt = state $ \s0 -> nextInt s0

-- Versão monádica do gerador de 3 números aleatórios
monadicNext3 :: State RNG [Int]
monadicNext3 = do
  x1 <- monadicNextInt
  x2 <- monadicNextInt
  x3 <- monadicNextInt
  return [x1, x2, x3]

-- Versão monádica do gerador de n números aleatórios
monadicNextN :: Int -> State RNG [Int]
monadicNextN n =
  mapM (const monadicNextInt) [1..n]


--------------------
-- Números primos --
--------------------

-- No caso de primos, o contexto que queremos guardar são os primos já
-- encontrados. Assim podemos usar esse contexto para calcular o
-- próximo primo

type PrimeGenState = (Int, [Int]) -- Último encontrado, primos conhecidos

initPrimes :: PrimeGenState
initPrimes = (3, [2,3])


pureNextPrime :: PrimeGenState -> (Int, PrimeGenState)
pureNextPrime (p, ps) =
  (p, nextState)
  where
    nextPrime = head [x | x <- [p + 2, p + 4 ..], prime x]
    nextState = (nextPrime, ps ++ [nextPrime]) -- a lista de primos é guardada
                                               -- na ordem inversa (não lazy!)
    prime x =
      all (\e -> x `mod` e /= 0) cands
      where
        lim = floor $ sqrt (fromIntegral x :: Float)
        cands = takeWhile (<= lim) ps

-- Usa o gerador como função pura
pureNext3Primes :: PrimeGenState -> ([Int], PrimeGenState)
pureNext3Primes s0 =
  let
    (v0, s1) = pureNextPrime s0 -- precisa carregar o estado
    (v1, s2) = pureNextPrime s1 -- entre cada uma das chamadas
    (v2, s3) = pureNextPrime s2 in
    ([v0, v1, v2], s3)

-- Usa o gerador como função pura
pureNextNPrimes :: Int -> PrimeGenState -> ([Int], PrimeGenState)
pureNextNPrimes 0 s0 = ([], s0)
pureNextNPrimes n s0 = (x : xs, sn)
  where
    -- Note que é necessário carregar o estado na mão
    (x, s1) = pureNextPrime s0
    (xs, sn) = pureNextNPrimes (n - 1) s1

-- -------------------------------------------------------------------
-- Exercício: Implemente a versão monádica do gerador de números
-- números primos utilizando a State Monad, a exemplo do que foi feito
-- acima para o gerador de números aleatórios.
-- -------------------------------------------------------------------


monadicNextPrime :: a
monadicNextPrime = undefined

monadicNext3Primes :: a
monadicNext3Primes = undefined

monadicNextNPrimes :: a
monadicNextNPrimes = undefined

main :: IO ()
main = do

  let r0 = simpleRandomInit 1234

  putStrLn "Aleatórios \"Puros\""
  print . fst $ pureNext3 r0
  print . fst $ pureNextN 3 r0

  putStrLn "Aleatórios \"Monadicos\""
  print $ evalState monadicNext3 r0
  print $ evalState (monadicNextN 3) r0

  let p0 = initPrimes

  putStrLn "Primos \"Puros\""
  print . fst $ pureNext3Primes p0
  print . fst $ pureNextNPrimes 3 p0

  putStrLn "Primos \"Monadicos\""
  -- Após implementar o exercício todas as anotações de tipos podem
  -- ser retiradas
  print (evalState monadicNext3Primes r0 :: [Int])
  print (evalState (monadicNextNPrimes (3 :: Int)) r0 :: [Int])
